/*
 * codecs.c,
 *
 * Copyright (C) 2004 Karl, Frankfurt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <ctype.h>

#include "../config.h"  /* autoconf output */
#include "app_data.h"
#include "codecs.h"

#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>

xvCodec tCodecs[NUMCODECS];
xvFFormat tFFormats[NUMCAPS];
xvAuCodec tAuCodecs[NUMAUCODECS];


const xvCodec none = {
    "NONE",
    "NONE",
    CODEC_NONE,
    CODEC_ID_NONE,
    0,
    0,
    NULL,
    NULL
};

const xvCodec mpeg1 = {
    "MPEG1",
    "MPEG 1",
    CODEC_MPEG1,
    CODEC_ID_MPEG1VIDEO,
    300000,
    24,
    NULL,
    "23.976|24|25|29.97|30|50|59.94|60"
};

const xvCodec mjpeg = {
    "MJPEG",
    "MJPEG",
    CODEC_MJPEG,
    CODEC_ID_MJPEG,
    300000,
    24,
    "7.5-30", // this is actually MPEG4 ... no idea if this is the same here
    NULL
};

const xvCodec mpeg4 = {
    "MPEG4",
    "MPEG 4 (DIVX)",
    CODEC_MPEG4,
    CODEC_ID_MPEG4,
    300000,
    24,
    "7.5-30",
    NULL
};

const xvCodec ms_div2 = {
    "MS_DIV2",
    "Microsoft DIVX 2",
    CODEC_MSDIV2,
    CODEC_ID_MSMPEG4V2,
    300000,
    24,
    "7.5-30", // this is actually MPEG4 ... no idea if this is the same here
    NULL
};

const xvCodec ms_div3 = {
    "MS_DIV3",
    "Microsoft DIVX 3",
    CODEC_MSDIV3,
    CODEC_ID_MSMPEG4V3,
    300000,
    24,
    "7.5-30", // this is actually MPEG4 ... no idea if this is the same here
    NULL
};

const xvCodec flv1 = {
    "FLASH_VIDEO",
    "Flash Video",
    CODEC_FLV,
    CODEC_ID_FLV1,
    300000,
    24,
    "7.5-30", // this is actually MPEG4 ... no idea if this is the same here
    NULL
};

const xvCodec dv = {
    "DV",
    "DV Video",
    CODEC_DV,
    CODEC_ID_DVVIDEO,
    300000,
    25,
    NULL, 
    "25|29.97"
};

const xvCodec mpeg2 = {
    "MPEG2",
    "MPEG2 Video",
    CODEC_MPEG2,
    CODEC_ID_MPEG2VIDEO,
    300000,
    24,
    NULL,
    "23.976|24|25|29.97|30|50|59.94|60"
};

const xvCodec svq1 = {
    "SVQ1",
    "S�rensen VQ 1",
    CODEC_SVQ1,
    CODEC_ID_SVQ1,
    300000,
    24,
    "7.5-30", // this is actually MPEG4 ... no idea if this is the same here
    NULL
};

// audio codecs

const xvAuCodec none_aucodec = {
    "NONE",
    "NONE",
    AU_CODEC_NONE,
    CODEC_ID_NONE
};

const xvAuCodec mp2 = {
    "MP2",
    "MPEG2",
    AU_CODEC_MP2,
    CODEC_ID_MP2
};

#ifdef HAVE_LIBMP3LAME
const xvAuCodec mp3 = {
    "MP3",
    "MPEG2 Layer 3",
    AU_CODEC_MP3,
    CODEC_ID_MP3
};
#endif //HAVE_LIBMP3LAME

const xvAuCodec pcm16 = {
    "PCM16",
    "PCM",
    AU_CODEC_PCM16,
    CODEC_ID_PCM_S16LE
};

// file formats

const xvFFormat none_format = {
    "NONE",
    "NONE",
    "NONE",
    NULL,
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat xwd = {
    "xwd",
    "X11 Window Dump",
    NULL,
    "xwd",
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat pnm = {
    "pnm",
    "Portable Anymap",
    NULL,
    "pnm|pgm|ppm",
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat png = {
    "png",
    "Portable Network Graphics",
    NULL,
    "png",
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat mng = {
    "mng",
    "Multiple Network Graphics",
    NULL,
    "mng",
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat jpg = {
    "jpg",
    "Joint Picture Expert Group",
    NULL,
    "jpg|jpeg",
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat qtr = {
    "qtr",
    "Quicktime Video",
    NULL,
    "mov",
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat qtj = {
    "qtj",
    "Quicktime Video",
    NULL,
    NULL,
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

const xvFFormat qtp = {
    "qtp",
    "Quicktime Video",
    NULL,
    NULL,
    CODEC_NONE,
    NULL,
    AU_CODEC_NONE,
    NULL,
    NULL
};

// from here on ffmpeg formats

const xvFFormat avi = {
    "avi",
    "Microsoft Audio Video Interleaved File",
    "avi",
    "avi",
    CODEC_MSDIV2,
    "MPEG1|MJPEG|MPEG4|MS_DIV2|MPEG2|DV",
    AU_CODEC_MP2,
#ifdef HAVE_LIBMP3LAME    
    "MP2|MP3|PCM16",
#else
    "MP2|PCM16",
#endif //HAVE_LIBMP3LAME
    avienc_init
};

const xvFFormat divx = {
    "divx",
    "General AVI file (DIVX default)",
    "avi",
    "mpeg|mpg",
    CODEC_MPEG4,
    "MPEG1|MJPEG|MPEG4|MS_DIV2|MPEG2|DV",
    AU_CODEC_MP2,
#ifdef HAVE_LIBMP3LAME    
    "MP2|MP3|PCM16",
#else
    "MP2|PCM16",
#endif //HAVE_LIBMP3LAME
    avienc_init
};

const xvFFormat asf = {
    "asf",
    "Microsoft Advanced Streaming Format",
    "asf",
    "asf",
    CODEC_MSDIV3,
    "MS_DIV3",
    AU_CODEC_NONE,
#ifdef HAVE_LIBMP3LAME    
    "MP2|MP3",
#else
    "MP2",
#endif //HAVE_LIBMP3LAME
    asf_init/*,
     guess_format */
};

const xvFFormat ff_flv1 = {
    "flv1",
    "Macromedia Flash Video Stream",
    "flv1",
    "flv|flv1",
    CODEC_FLV,
    "FLASH_VIDEO",
    AU_CODEC_NONE,
    NULL,
    flvenc_init/*,
     guess_stream_format */
};

const xvFFormat swf = {
    "swf",
    "Macromedia Shockwave Flash File",
    "swf",
    "swf",
    CODEC_FLV,
    "FLASH_VIDEO|MJPEG",
    AU_CODEC_MP3,
#ifdef HAVE_LIBMP3LAME    
    "MP2|MP3",
#else
    "MP2",
#endif //HAVE_LIBMP3LAME
    swf_init/*,
     guess_stream_format */
};

const xvFFormat dv_format = {
    "dv",
    "DV Video Format",
    "dv",
    "dv",
    CODEC_DV,
    "DV",
    AU_CODEC_MP2,
    "PCM16",
    ff_dv_init/*,
     guess_stream_format */
};

const xvFFormat mpeg = {
    "mpeg",
    "MPEG1 System Format",
    "mpeg",
    "m1v|vcd",
    CODEC_MPEG1,
    "MPEG1",
    AU_CODEC_MP2,
    "MP2",
    mpegps_init/*,
     guess_stream_format */
};

const xvFFormat svcd = {
    "mpeg2",
    "MPEG2 PS Format",
    "svcd",
    "m2v|svcd",
    CODEC_MPEG2,
    "MPEG2",
    AU_CODEC_MP2,
    "MP2",
    mpegps_init/*,
     guess_stream_format */
};

const xvFFormat mov = {
    "mov",
    "Quicktime Format",
    "mov",
    "mov|qt",
    CODEC_MPEG4,
    "MPEG4|SVQ1|DV",
    AU_CODEC_MP2,
#ifdef HAVE_LIBMP3LAME    
    "MP2|MP3|PCM16",
#else
    "MP2|PCM16",
#endif //HAVE_LIBMP3LAME
    mpegps_init/*,
     guess_stream_format */
};

/*
 * must be called by main() to initialize data structures
 */
void xv_codecs_init () {
    int i = 0;
    //const xvCodec tCodecs[] = { none, mpeg1, mjpeg, mpeg4, ms_div2, ms_div3, flv1 };
    //const xvFFormat tFFormats[] = { avi, mpeg, asf, ff_flv1 };
    
    // these are absolutely required to be in the same order as the
    // enumeration of CAP_* types in job.h
    tCodecs[i++] = none;
#ifdef HAVE_LIBAVCODEC
#ifdef HAVE_LIBAVFORMAT
    tCodecs[i++] = mpeg1;
    tCodecs[i++] = mjpeg;
    tCodecs[i++] = mpeg4;
    tCodecs[i++] = ms_div2;
    tCodecs[i++] = ms_div3;
    tCodecs[i++] = flv1;
    tCodecs[i++] = dv;
    tCodecs[i++] = mpeg2;
    tCodecs[i++] = svq1;
#endif //HAVE_LIBAVFORMAT
#endif //HAVE_LIBAVCODEC

        // dto.
        // the target item of the array will be found by CAP_XXX - CAP_FFM
        i = 0;
    tFFormats[i++] = none_format;
    tFFormats[i++] = xwd;
    tFFormats[i++] = pnm;
#ifdef HAVE_LIBPNG
        tFFormats[i++] = png;
    tFFormats[i++] = mng;
#endif //HAVE_LIBPNG
#ifdef HAVE_LIBJPEG
        tFFormats[i++] = jpg;
#endif //HAVE_LIBJPEG
#ifdef WITHGT
        tFFormats[i++] = qtr;
#ifdef HAVE_LIBJPEG
        tFFormats[i++] = qtj;
#endif //HAVE_LIBJPEG
#ifdef HAVE_LIBPNG
        tFFormats[i++] = qtp;
#endif //HAVE_LIBPNG
#endif //WITHGT
#ifdef HAVE_LIBAVCODEC
#ifdef HAVE_LIBAVFORMAT
    tFFormats[i++] = avi;
    tFFormats[i++] = divx;
    tFFormats[i++] = asf;
    tFFormats[i++] = ff_flv1;
    tFFormats[i++] = swf;
    tFFormats[i++] = dv_format;
    tFFormats[i++] = mpeg;
    tFFormats[i++] = svcd;
    tFFormats[i++] = mov;
#endif //HAVE_LIBAVFORMAT
#endif //HAVE_LIBAVCODEC

        // audio codecs
    i = 0;
    tAuCodecs[i++] = none_aucodec;
#ifdef HAVE_LIBAVCODEC
#ifdef HAVE_LIBAVFORMAT
    tAuCodecs[i++] = mp2;
#ifdef HAVE_LIBMP3LAME
    tAuCodecs[i++] = mp3;
#endif //HAVE_LIBMP3LAME
    tAuCodecs[i++] = pcm16;
#endif //HAVE_LIBAVFORMAT
#endif //HAVE_LIBAVCODEC

}

/*
 * find ffmpeg codec id from xvidcap's
 * 0 is a valid id ... therefore we return -1 on failure
 */
int xv_trans_codec (int xv_codec) {
    int i, ret = -1;
    
    for (i=0; i < NUMCODECS; i++) {
        if (tCodecs[i].id == xv_codec) ret = tCodecs[i].ffmpeg_id;
    }
    return ret;
}

/*
 * check if string is element in list (e.g. allowed_vid_codecs)
 */
int xv_is_element (char *xvList, char *xvElement) {
    char llist[512];
    char lelement[512];
    char *found = NULL;
    //    char *llist = NULL, *lelement = NULL;
    int ret, y;
    
    sprintf (llist, "|%s|", xvList);
    for (y = 0; y < strlen (llist); y++) {
        llist[y] = toupper (llist[y]);
    }
    
    sprintf (lelement, "|%s|", xvElement);
    for (y = 0; y < strlen (lelement); y++) {
        lelement[y] = toupper (lelement[y]);
    }
    
    found = strstr ( llist, lelement );
    if ( found != NULL ) ret = 1;
    else ret = 0;
    return ret;
    //	return 0;
}

/*
 * enumerate list (e.g. allowed_vid_codecs)
 */
char* xv_next_element (char* list) {
    static char *p_list;
    static char llist[512];
    char *ret;
    
    if ( list != NULL ) {
        sprintf (llist, "|%s|", list);
        ret = strtok_r (llist, "|", &p_list);
    } else {
        ret = strtok_r (NULL, "|", &p_list);
    }
    return ret;
}

/*
 * find target based on filename
 * returns CAP_* or 0 if none found
 */
int xv_codec_get_target_from_filename (char *file) {
    char *ext = NULL, *element = NULL;
    int ret = 0, n;
    
    ext = rindex (file, '.');
    if ( ext == NULL ) {
        return ret;
    }
    ext++;
    
    for (n = CAP_NONE; n < NUMCAPS; n++) {
        if (tFFormats[n].extensions) {
            element = xv_next_element (tFFormats[n].extensions);
            while ( element != NULL ) {
                if ( strcasecmp (ext, element) == 0 ) {
                    // then we have found the right extension in target n
                    ret = n;
                    return ret;
                }
                element = xv_next_element (NULL);
            }
        }
    }
    
    return ret;
}

/*
 * check if fps rate is valid for given codec
 * returns 0 for false or 1 for true
 */
int xv_codec_is_valid_fps (int fps, int codec) {
//    struct lconv *loc = NULL;
    char *element = NULL;
//    int ret = 0, rep_dp = 0;
    
//    printf ("fps: %f - codec: %i %s\n", fps, codec, tCodecs[codec].name);
//    printf ("fps: %s\nfps_ranges: %s\n", tCodecs[codec].allowed_fps, tCodecs[codec].allowed_fps_ranges);
/*    loc = localeconv ();
    if (! loc) {
        fprintf (stderr, "Cant' determine valid fps due to failure to determine current locale.\n");
        return 0;
    } */
    //    printf ("this locale's decimal point is: %s\n", loc->decimal_point);
//    if ( strcasecmp ( loc->decimal_point, "." ) != 0 ) rep_dp = 1;
    
    if (tCodecs[codec].allowed_fps) {
        element = xv_next_element (tCodecs[codec].allowed_fps);
        while (element) {
            int f;
/*            float f = (float) -1;
            // the conversion is always locale dependant ... so we will need
            // to replace the decimal point in the codec definition by whatever is
            // the locale's decimal point
            if (rep_dp) {
                char *dot = index (element, '.');
                if (dot) *dot = *loc->decimal_point;
                //            printf ("changed element: %s\n", element);
            }
            sscanf (element, "%f", &f); */
            f = get_int_from_float_string(element);
            if ( f < 0 ) fprintf (stderr, "Non-fatal error in the definition of valid fps for codec %s\n",
            tCodecs[codec].name);
            else if ( f == fps ) return 1;
            element = xv_next_element (NULL);
        }
    }

    if (tCodecs[codec].allowed_fps_ranges) {
        element = xv_next_element (tCodecs[codec].allowed_fps_ranges);
        while (element) {
//            double start = -1, end = -1;
//            char *endelement;
            // the conversion is always locale dependant ... so we will need
            // to replace the decimal point in the codec definition by whatever is
            // the locale's decimal point
/*            if (rep_dp) {
                char *dot;
                dot = index (element, '.');
                while (dot) {
                    *dot = *loc->decimal_point;
                    //                printf ("changed element: %s\n", element);
                    dot = index (element, '.');
                }
            } */
            int start = 0, end = 0;
            char *start_str = strdup(element);
            char *ptr = index( start_str, '-' );
            *ptr = '\0';
            ptr++;
            
/*            start = strtod (element, &endelement);
            endelement++;
            end = strtod (endelement, NULL);
            //        printf ("start: %f - end: %f - orig: %s - fps: %f\n", start, end, element, fps);
*/
            start = get_int_from_float_string(start_str);
            end = get_int_from_float_string(ptr);

            if ( start < 0 || end < 0 ) {
                fprintf (stderr, "Non-fatal error in the definition of valid fps ranges for codec %s\n",
                tCodecs[codec].name);
            }
            else if ( start <= fps && fps <= end ) return 1;
            element = xv_next_element (NULL);
        }
    }
    
    //FIXME: what kind of error do I want to report with the return code?
    return 0;
}


/*
 * count elements in a list (e.g. allowed_vid_codecs)
 */
int xv_num_elements (char* list) {
    static char *p_list;
    static char llist[512];
    char *elem;
    int elem_count = 0;
    
    if ( list != NULL ) {
        sprintf (llist, "|%s|", list);
        elem = strtok_r (llist, "|", &p_list);
        do {
            elem_count++;
            elem = strtok_r (NULL, "|", &p_list);
        } while ( elem != NULL );
    } 
    return elem_count;
}

