/*
 * main.c,
 *
 * Copyright (C) 1997,98 Rasca, Berlin
 * Copyright (C) 2003,04 Karl, Frankfurt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../config.h"  /* autoconf output */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>
#include <ctype.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#ifdef HAVE_SHMAT
#include <X11/extensions/XShm.h>
#endif // HAVE_SHMAT
#ifdef HasDGA
#include <X11/extensions/xf86dga.h>
#endif //HasDGA
#ifdef HasVideo4Linux
#include <video.h>
#endif //HasVideo4Linux
#include "main.h"
#include "control.h"
#include "frame.h"
#include "job.h"
#include "codecs.h"


/* GLOBAL VARIABLES */
AppData sapp, *app;
extern xvCodec tCodecs[NUMCODECS];
extern xvFFormat tFFormats[NUMCAPS];
extern xvAuCodec tAuCodecs[NUMAUCODECS];

/* HELPER FUNCTIONS */
void
usage (char *prog) {
    int n, m;
    char *element;
    
    printf ("Usage: %s, ver %s, (c) rasca, berlin 1997,98,99, khb (c) 2003,04\n", prog, VERSION);
    printf ("[--fps #.#]         frames per second (float)\n");
    printf ("[--verbose #]       verbose level, '-v' is --verbose 1\n");
    printf ("[--time #.#]        time to record in seconds (float)\n");
    printf ("[--frames #]        frames to record, don't use it with --time\n");
    printf ("[--continue [yes|no]] 	autocontinue after maximum frames/time\n");
    printf ("[--cap_geometry #x#[+#+#]]  size of the capture window (WIDTHxHEIGHT+X+Y)\n");
    printf ("[--start_no #]      start number for the file names\n");
    #ifdef HAVE_LIBZ
    printf ("[--compress <0-9>]  save the frames compressed, '.gz' is added\n");
    #endif
    printf ("[--use_clone]       use the CLON chunk for MNGs if needed\n");
    printf ("[--expand_to_24bit] expand 15/16bpp to 24bpp (only for PPM)\n");
    #ifdef HAVE_LIBJPEG
    printf ("[--quality #]       quality value for JPEG files (100 - 0)\n");
    #endif
    printf ("[--source <src>]    select input source: x11, shm\n");
    printf ("[--file <file>]     file pattern, e.g. out%%03d.xwd\n");
    printf ("[--gui [yes|no]]    turn on/off gui\n");
    printf ("[--sf|--mf]         request single-frame or multi-frame capture mode\n");
    printf ("[--auto]            cause auto-detection of output format, video-, and audio codec\n");
    printf ("[--codec <codec>]   specify codec to use for multi-frame capture\n");
    printf ("[--codec-help]      list available codecs for multi-frame capture\n");
    printf ("[--format <format>] specify file format to override the extension in the filename\n");
    printf ("[--format-help]     list available file formats\n");
#ifdef HAVE_FFMPEG_AUDIO
    printf ("[--aucodec <codec>] specify audio codec to use for multi-frame capture\n");
    printf ("[--aucodec-help]    list available audio codecs for multi-frame capture\n");
    printf ("[--audio [yes|no]]  turn on/off audio capture\n");
    printf ("[--audio_in <src>]  specify audio input device or '-' for pipe input\n");
    printf ("[--audio_rate #]    sample rate for audio capture\n");
    printf ("[--audio_bits #]    bit rate for audio capture\n");
    printf ("[--audio_channels #]    number of audio channels\n");
#endif // HAVE_FFMPEG_AUDIO

    printf ("Supported output formats:\n");
    for ( n = CAP_NONE; n < NUMCAPS; n++) {
        if (tFFormats[n].extensions) {
            printf (" %s", tFFormats[n].longname);
            for (m = strlen (tFFormats[n].longname); m < 40; m++) printf (" ");
            printf ("(");
            element = xv_next_element (tFFormats[n].extensions);
            while ( element != NULL ) {
                printf (".%s", element);
                element = xv_next_element (NULL);
                if ( element != NULL ) printf (", ");
            }
            printf (")\n");
        }
    }
exit (1);
}


void
cleanup_when_interrupted (int signal) {
    Job *job = job_ptr ();
    
    XVC_StopCapture (job);
}


/*
 * main()
 */
int
main (int argc, char **argv) {
    
    //
    // local variables
    //
    struct option options[] = {
        {"fps",      required_argument, NULL, 0},
        {"file",     required_argument, NULL, 0},
        {"verbose",  required_argument, NULL, 0},
        {"time",     required_argument, NULL, 0},
        {"frames",   required_argument, NULL, 0},
        {"cap_geometry", required_argument, NULL, 0},
        {"start_no", required_argument, NULL, 0},
        {"compress", required_argument, NULL, 0},
        {"use_clone", no_argument,		NULL, 0},
        {"expand_to_24bit", no_argument, NULL, 0},
        {"quality",  required_argument, NULL, 0},
        {"bpp", 	 required_argument, NULL, 0},
        {"source", 	 required_argument, NULL, 0},
        {"sync", 	 no_argument,		NULL, 0},
        {"step", 	 required_argument,	NULL, 0},
        {"gui",     optional_argument, NULL, 0},
        {"audio",     optional_argument, NULL, 0},
        {"audio_in", required_argument, NULL, 0},
        {"audio_rate", required_argument, NULL, 0},
        {"audio_bits", required_argument, NULL, 0},
        {"audio_channels", required_argument, NULL, 0},
        {"continue", optional_argument, NULL, 0},
        {"sf", no_argument,		NULL, 0},
        {"mf", no_argument,		NULL, 0},
        {"codec", 	 required_argument, NULL, 0},
        {"codec-help", no_argument,		NULL, 0},
        {"format", 	 required_argument, NULL, 0},
        {"format-help", no_argument,		NULL, 0},
        {"aucodec", 	 required_argument, NULL, 0},
        {"aucodec-help", no_argument,		NULL, 0},
        {"auto", no_argument, NULL, 0 },
        {NULL, 0, NULL, 0},
    };
    
    
    int opt_index = 0, c, resultCode; //, flags =0;
    CapTypeOptions s_tmp_capture_options, *tmp_capture_options, *target;
    int current_mode_by_filename = -1, current_mode_by_target = -1;
    xvErrorListItem *errors_after_cli = NULL;
    
    #ifdef HasDGA
    int dga_evb, dga_errb;
    #endif //HasDGA
    
    // FIXME: this needs to be unconditional and the condition must be
    // verified within the init function
    #ifdef HAVE_LIBAVCODEC
    #ifdef HAVE_LIBAVFORMAT
    xv_codecs_init ();
    #endif // HAVE_LIBAVFORMAT
    #endif // HAVE_LIBAVCODEC
    xv_errors_init ();
    
    
    //
    // some variable initialization
    //
    app = &sapp;
    app_data_init (app);
    app_data_set_defaults (app);
    tmp_capture_options = &s_tmp_capture_options;
    cap_type_options_init (tmp_capture_options);
    
    
    // this is a hook for a GUI to do some pre-init functions ...
    // possibly to set some fallback options read from a rc file or Xdefaults
    if (! XVC_PreInit ( argc, argv ) ) {
        fprintf (stderr, "Can't do GUI pre_init ... aborting\n");
        exit (2);
    }
    
    /* read options file now */
    XVC_ReadOptionsFile ( app );
    
    // FIXME: add all sorts of new cli options
    while ((c = getopt_long (argc, argv, "v", options, &opt_index)) != -1) {
        switch (c) {
            case 0: /* is a long option */
                switch (opt_index) {
                    case 0: /* frames per second */
                        tmp_capture_options->fps = get_int_from_float_string(optarg);
//                        printf("read: %i\n", (int) tmp_capture_options->fps);
                        break;
                    case 1: /* file for saving */
                        tmp_capture_options->file = strdup (optarg);
                        break;
                    case 2:
                        app->verbose = atoi (optarg);
                        break;
                    case 3: /* time to record */
                        tmp_capture_options->time = atoi (optarg);
//                        if ( tmp_capture_options->time != 0 ) tmp_capture_options->frames = 0;
                        break;
                    case 4:
                        tmp_capture_options->frames = atoi (optarg);
//                        if ( tmp_capture_options->frames != 0 ) tmp_capture_options->time = 0;
                        break;
                    case 5: /* cap_geometry */
                        sscanf (optarg, "%dx%d+%d+%d", &(app->cap_width), &(app->cap_height), &(app->cap_pos_x), &(app->cap_pos_y));
                        break;
                    case 6: // start_no
                        tmp_capture_options->start_no = atoi (optarg);
                        break;
                    case 7: // compress
                        tmp_capture_options->compress = atoi (optarg);
                        break;
                    case 8: // use_mng_clone
                        app->flags |= FLG_MNG_CLONE;
                        break;
                    case 9: // expand to 24 bits
                        app->flags |= FLG_EXPAND_TO_24BIT;
                        break;
                    case 10: // quality
//                        if ((atoi (optarg) < 0) || (atoi (optarg) > 100))
                            tmp_capture_options->quality = atoi (optarg);
                        break;
                    case 11:
                        tmp_capture_options->bpp = atoi (optarg);
                        break;
                    case 12: // source
                        app->source = strdup (optarg);
                        /*                        if (strstr(optarg, "v4l") != NULL) {
                         app->flags &= ~FLG_SOURCE;
                         app->flags |= FLG_USE_V4L;
                         if (strchr(optarg, ':') != NULL) {
                         app->device = strchr(optarg, ':')+1;
                         }
                         } else if (strstr(optarg, "dga") != NULL) {
                         app->flags &= ~FLG_SOURCE;
                         app->flags |= FLG_USE_DGA;
                         } else if (strstr(optarg, "shm") != NULL) {
                         app->flags &= ~FLG_SOURCE;
                         app->flags |= FLG_USE_SHM;
                         } else (
                         app->flags &= ~FLG_SOURCE;
                         printf("using normal x server access ..\n");
                         } */
                        break;
                    case 13:
                        app->flags |= FLG_SYNC;
                        break;
                    case 14:
                        tmp_capture_options->step = atoi (optarg);
                        break;
                    case 15: // gui
                    {
                        char *tmp;
                        if (! optarg) {
                            if (optind < argc) {
                                tmp = (argv[optind][0] == '-')?"yes":argv[optind++];
                            } else {
                                tmp = "yes";
                            }
                        } else {
                            tmp = strdup (optarg);
                        }
                        if (strstr (tmp, "no") != NULL) {
                            app->flags |= FLG_NOGUI;
                        } else {
                            app->flags &= ~FLG_NOGUI;
                        }
                    }
                    break;
                    case 16: //audio
                    {
                        char *tmp;
                        if (! optarg) {
                            if (optind < argc) {
                                tmp = (argv[optind][0] == '-')?"yes":argv[optind++];
                            } else {
                                tmp = "yes";
                            }
                        } else {
                            tmp = strdup (optarg);
                        }
                        if (strstr (tmp, "no") != NULL) {
                            tmp_capture_options->audioWanted = 0;
                        } else {
                            tmp_capture_options->audioWanted = 1;
                        }
                    }
                    break;
                    case 17: //audio_in
                        tmp_capture_options->audioWanted = 1;
                        app->snddev = strdup (optarg);
                        break;
                    case 18: //audio_rate
                        tmp_capture_options->audioWanted = 1;
                        tmp_capture_options->sndrate = atoi (optarg);
                        break;
                    case 19: //audio_bits
                        tmp_capture_options->audioWanted = 1;
                        tmp_capture_options->sndsize = atoi (optarg);
                        break;
                    case 20: //audio_channels
                        tmp_capture_options->audioWanted = 1;
                        tmp_capture_options->sndchannels = atoi (optarg);
                        break;
                    case 21: //continue
                    {
                        char *tmp;
                        if (! optarg) {
                            if (optind < argc) {
                                tmp = (argv[optind][0] == '-')?"yes":argv[optind++];
                            } else {
                                tmp = "yes";
                            }
                        } else {
                            tmp = strdup (optarg);
                        }
                        if (strstr (tmp, "no") != NULL) {
                            app->flags &= ~FLG_AUTO_CONTINUE;
                        } else {
                            app->flags |= FLG_AUTO_CONTINUE;
                        }
                    }
                    break;
                    case 22: // single-frame
                        app->current_mode = 0;
                        break;
                    case 23: // multi-frame
                        app->current_mode = 1;
                        break;
                    case 24: // codec
                    {
                        int n, m = 0;
                        //				char tmp_codec[512];
                        Boolean found = FALSE;
                        
                        for (n=1; n < NUMCODECS; n++) {
                            if (strcasecmp (optarg, tCodecs[n].name) == 0) {
                                found = TRUE;
                                m = n;
                            }
                        }
                        if ( strcasecmp (optarg, "auto") == 0 ) {
                            found = TRUE;
                            m = 0; // set to CODEC_NONE which will get overridden later
                        }
                        
                        if (found) {
                            tmp_capture_options->targetCodec = m;
                        } else {
                            usage (argv[0]);
                        }
                    }
                    break;
                    case 25: // codec-help
                    {
                        int n, m;
                        char tmp_codec[512];
                        
                        printf ("%s, ver %s, (c) rasca, berlin 1997,98,99, khb (c) 2003,04\n", argv[0], VERSION);
                        printf ("Available codecs for multi-frame capture: ");
                        
                        for (n=1; n<NUMCODECS; n++) {
                            for (m=0; m < strlen (tCodecs[n].name); m++) {
                                tmp_codec[m] = tolower (tCodecs[n].name[m]);
                            }
                            tmp_codec[strlen (tCodecs[n].name)] = '\0';
                            printf ("%s", tmp_codec);
                            if ( n < (NUMCODECS-1) ) printf (", ");
                        }
                        printf ("\n");
                        printf ("Specify 'auto' to use the file format's default codec.\n");
                        exit (1);
                    }
                    break;
                    case 26: // format
                    {
                        int n, m = -1;
                        char tmp_format[512], *element;
                        Boolean found = FALSE;
                        
                        for ( n = CAP_NONE; n < NUMCAPS; n++) {
                            if (tFFormats[n].extensions) {
                                element = xv_next_element (tFFormats[n].extensions);
                                while ( element != NULL ) {
                                    if (strcasecmp (optarg, element)==0) {
                                        found = TRUE;
                                        sprintf (tmp_format, "bla.%s", element);
                                        m = xv_codec_get_target_from_filename (tmp_format);
                                    }
                                    element = xv_next_element (NULL);
                                }
                            }
                        }
                        if (strcasecmp (optarg, "auto")==0) {
                            found = TRUE;
                            m = 0;
                        }
                        
                        if (found) {
                            if (m > -1) {
                                tmp_capture_options->target = m;
                                // if target is auto-detected, we also auto-detect targetCodec and au_targetCodec
/*                                if ( m == 0 ) { 
                                    tmp_capture_options->targetCodec = 0;
                                    tmp_capture_options->au_targetCodec = 0;
                                } */
                            } else {
                                fprintf (stderr, "Specified file format '%s' not supported with this binary.", optarg);
                                fprintf (stderr, "Resetting to file format auto-detection.\n");
                                tmp_capture_options->target = 0;
                            }
                        } else {
                            fprintf (stderr, "Unknown file format '%s' specified.\n", optarg);
                            usage (argv[0]);
                        }
                    }
                    break;
                    case 27: // format-help
                    {
                        int n, m;
                        char tmp_format[512];
                        
                        printf ("%s, ver %s, (c) rasca, berlin 1997,98,99, khb (c) 2003,04\n", argv[0], VERSION);
                        printf ("Available file formats: ");
                        
                        for ( n = CAP_NONE; n < NUMCAPS; n++) {
                            if (tFFormats[n].extensions) {
                                for (m=0; m < strlen (tFFormats[n].name); m++) {
                                    tmp_format[m] = tolower (tFFormats[n].name[m]);
                                }
                                tmp_format[strlen (tFFormats[n].name)] = '\0';
                                printf ("%s", tmp_format);
                                if ( n < (NUMCAPS-1) ) printf (", ");
                            }
                        }
                        printf ("\n");
                        printf ("Specify 'auto' to force file format auto-detection.\n");
                        exit (1);
                    }
                    break;
                    case 28: // aucodec
                    {
                        int n, m = 0;
                        //				char tmp_codec[512];
                        Boolean found = FALSE;
                        
                        for (n=1; n < NUMAUCODECS; n++) {
                            if (strcasecmp (optarg, tAuCodecs[n].name) == 0) {
                                found = TRUE;
                                m = n;
                            }
                        }
                        if ( strcasecmp (optarg, "auto") == 0 ) {
                            found = TRUE;
                            m = 0; // set to CODEC_NONE which will get overridden later
                        }
                        
                        if (found) {
                            tmp_capture_options->au_targetCodec = m;
                        } else {
                            usage (argv[0]);
                        }
                    }
                        break;
                    case 29: // aucodec-help
                    {
                        int n, m;
                        char tmp_codec[512];
                        
                        printf ("%s, ver %s, (c) rasca, berlin 1997,98,99, khb (c) 2003,04\n", argv[0], VERSION);
                        printf ("Available audio codecs for multi-frame capture: ");
                        
                        for (n=1; n<NUMAUCODECS; n++) {
                            for (m=0; m < strlen (tAuCodecs[n].name); m++) {
                                tmp_codec[m] = tolower (tAuCodecs[n].name[m]);
                            }
                            tmp_codec[strlen (tAuCodecs[n].name)] = '\0';
                            printf ("%s", tmp_codec);
                            if ( n < (NUMAUCODECS-1) ) printf (", ");
                        }
                        printf ("\n");
                        printf ("Specify 'auto' to use the file format's default audio codec.\n");
                        exit (1);
                    }
                        break;
                    case 30: // auto
                        tmp_capture_options->target = 0;
                        tmp_capture_options->targetCodec = 0;
                        tmp_capture_options->au_targetCodec = 0;
                        break;
                    default:
                        usage (argv[0]);
                        break;
                }
                break;
                
            case 'v':
                app->verbose++;
                break;
                
            default:
                usage (argv[0]);
                break;
        }
    }
    if (argc != optind)
        usage (argv[0]);
    
    
    // evaluate tmp_capture_options and set real ones accordingly
    // determine capture mode
    //
    // either explicitly set in cli options
    // or default_mode in options (prefs/xdefaults), the latter does not need
    // special action because it will have been set by now or fall back to the
    // default set in init_app_data.
    
    // app->current_mode is either -1 (not explicitly specified), 0 (single_frame),
    // or 1 (multi_frame) ... setting current_mode_by_* accordingly
    
    // by filename
    if ( tmp_capture_options->file == NULL ) current_mode_by_filename = -1;
    else {
        current_mode_by_filename = xv_codec_get_target_from_filename (tmp_capture_options->file);
        if ( current_mode_by_filename >= CAP_FFM ) current_mode_by_filename = 1;
        else if ( current_mode_by_filename > 0 ) current_mode_by_filename = 0;
        // if we can't determine capture type by filename treat as not
        // set for all practical purposes and leave error handling to
        // error handling
        else current_mode_by_filename = -1;
    }
    
    // by target
    if ( tmp_capture_options->target < 0 ) current_mode_by_target = -1;
    else {
        if ( tmp_capture_options->target >= CAP_FFM ) current_mode_by_target = 1;
        else if ( tmp_capture_options->target > 0 ) current_mode_by_target = 0;
        // if we have an invalide target treat as not
        // set for all practical purposes and leave error handling to
        // error handling
        else current_mode_by_target = -1;
    }
    
    if ( app->current_mode >= 0 ) {
        if (app->current_mode == 0) target = &(app->single_frame);
        else target = &(app->multi_frame);
    } else {
        app->current_mode = app->default_mode;
        // in case default_mode is not set correctly (will trigger error later)
        if ( app->current_mode < 0 || app->current_mode > 1 ) app->current_mode = 0;
        // preferences defaults are overridden by filename
        if (current_mode_by_filename >= 0) app->current_mode = current_mode_by_filename;
        // filename is overridden by format
        if (current_mode_by_target >= 0) app->current_mode = current_mode_by_target;
        
        if (app->current_mode == 0) target = &(app->single_frame);
        else target = &(app->multi_frame);
    }
    printf ("Doing %s capture.\n", ((app->current_mode == 0)?"single-frame":"multi-frame"));
    
    // merge cli options with app
    if (merge_cap_type_and_app_data (tmp_capture_options, app) < 1) {
        fprintf (stderr, "main: Unrecoverable error while merging options, please contact the xvidcap project.\n");
        exit (1);
    }
    
    // get erros in options to pass to XVC_InitGUI
    {
        int rc;
            rc = 0;
            errors_after_cli = app_data_validate (app, 1, &rc);
            if (rc == -1) {
                fprintf (stderr, "main: Unrecoverable error while validating options, please contact the xvidcap project.\n");
                exit (1);
            } 
    } 
    
    // these are the hooks for a GUI to create the GUI,
    // the selection frame, and do some initialization ...
    if ( ! XVC_CreateGUI () ) {
        fprintf (stderr, "Can't create GUI ... aborting\n");
        exit (2);
    }
    if ( ! XVC_CreateFrame () ) {
        fprintf (stderr, "Can't create selection Frame ... aborting\n");
        exit (2);
    }
    if ( ! XVC_InitGUI (errors_after_cli) ) {
        fprintf (stderr, "Can't initialize GUI ... aborting\n");
        exit (2);
    }
    if ( app->cap_pos_x >= 0 || app->cap_pos_y >= 0 ) {
        XVC_ChangeFrame ( app->cap_pos_x, app->cap_pos_y, app->cap_width, app->cap_height, TRUE );
    }
    
    
    // FIXME: make the options output take new app_data structure with
    // parallel capTypeOptions into account
    if (app->verbose) {
        char *mp;
        switch (app->mouseWanted) {
            case 2:
                mp = "black";
                break;
            case 1:
                mp = "white";
                break;
            default:
                mp = "none";
        }
        
        printf ("Current settings:\n");
        printf ("  flags             = %d\n", app->flags);
        printf ("  capture mode      = %s\n", ((app->current_mode == 0) ? "single-frame":"multi-frame"));
        printf ("  position	    = %ix%i", app->cap_width, app->cap_height);
        if (app->cap_pos_x >= 0) printf ("+%i+%i", app->cap_pos_x, app->cap_pos_y);
        printf ("\n");
        printf ("  frames per second = %.2f\n", ((float) target->fps/100));
        printf ("  file pattern      = %s\n", target->file);
        printf ("  file format       = %s\n", ((target->target == CAP_NONE)?"AUTO":tFFormats[target->target].longname));
        printf ("  video encoding    = %s\n", ((target->targetCodec == CODEC_NONE)?"AUTO":tCodecs[target->targetCodec].name));
        printf ("  audio codec       = %s\n", ((target->au_targetCodec == CODEC_NONE)?"AUTO":tAuCodecs[target->au_targetCodec].name));
        printf ("  verbose level     = %d\n", app->verbose);
        printf ("  frames to store   = %d\n", target->frames);
        printf ("  time to capture   = %i sec\n", target->time);
        printf ("  autocontinue      = %s\n", ((app->flags & FLG_AUTO_CONTINUE)?"yes":"no"));
        #ifdef HAVE_LIBZ
        printf ("  compression level = %d\n", target->compress);
        #endif
        #ifdef HAVE_LIBJPEG
        printf ("  quality (jpeg)    = %d\n", target->quality);
        #endif
        printf ("  input source      = %s (%d)\n", app->source, app->flags & FLG_SOURCE);
        printf ("  capture pointer   = %s\n", mp);
#ifdef HAVE_FFMPEG_AUDIO
        printf ("  capture audio     = %s\n", ((target->audioWanted == 1)?"yes":"no"));
        printf ("   - input          = %s\n", app->snddev);
        printf ("   - sample rate    = %i\n", target->sndrate);
        printf ("   - bit rate       = %i\n", target->sndsize);
        printf ("   - channels       = %i\n", target->sndchannels);
#endif // HAVE_FFMPEG_AUDIO
        printf ("  animate command   = %s\n", target->play_cmd);
        printf ("  make video command= %s\n", target->video_cmd);
        printf ("  edit frame command= %s\n", target->edit_cmd);
        printf ("  help command      = %s\n", app->help_cmd);
    }
    
    signal (SIGINT, cleanup_when_interrupted);
    
    // this is a hook for the GUI's main loop
    resultCode = XVC_RunGUI ();
    return (resultCode);
}

